/*
    Copyright (C) 2002-2004 Robert Muth <robert at muth dot org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of June 1991.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program in form of the file COPYING; 
    if not, write to the 

    Free Software Foundation, Inc. <http://www.fsf.org>
    59 Temple Place, Suite 330, 
    Boston, MA 02111-1307  USA

    $Id: basic.C,v 1.5 2004/04/24 15:37:12 muth Exp $
*/

/* ================================================================== */


#include <iostream>
#include <string>
#include <sstream>
#include <iomanip>

#include "basic.H"

#include <execinfo.h>

#include <signal.h>

#include <stdlib.h>


/* ================================================================== */

static char *image_name;
static char disabled[256];

/* ================================================================== */

VOID DisableMessage(UINT32 prefix)
{
    ASSERTX( prefix < 256 );
    
    disabled[prefix] = 1;
}

/* ================================================================== */

VOID EnableMessage(UINT32 prefix)
{
    ASSERTX( prefix < 256 );

    disabled[prefix] = 0;
}

/* ================================================================== */

VOID Message(string prefix, string message, BOOL dump_trace, BOOL terminate)
{
    if( !disabled[UINT32(prefix[0])] )
    {
        UINT32 len = message.length();
    
        if( len > 1 )
        {
            len--;
            
            while( --len )
            {
                if( message[len] == '\n' )
                {
                message.insert(len+1,prefix);
                }
            }
        }
  
        message.insert(0,prefix);  

        
        cout << message;
    }
    
    if( dump_trace )
    {
        DumpTrace();
    }
    
    if ( terminate )
    {
        abort();
    }
    
}

/* ================================================================== */
static BOOL DoTrace = 1;

/* ================================================================== */
VOID DisableDumpTrace()
{
    DoTrace = 0;
}

/* ================================================================== */

VOID DumpTrace()
{
    if( !DoTrace ) return;
    DoTrace = 0;
    
    // Store up to SIZE return address of the current program state in
    // ARRAY and return the exact number of values stored. 
    
    const int size = 128;
    void *array[size];

    int actual_size = backtrace(array, size);
    
    // Return names of functions from the backtrace list in ARRAY in a newly
    // allocated memory block.  

    string prefix("T: ");


    char **trace  = backtrace_symbols(array,size);
    
    string suffix("\n");
    
    
    for( int i=0; i < actual_size; i++ )
    {
        cout << i << array[i] << trace[i] << endl;
        Message( prefix, trace[i],0,0);
    }

    ostringstream commandline;
    
    commandline << "addr2line -C -f -e ";
    commandline << image_name << hex;
    
    for( int i=1; i < actual_size; i++ )
    {
        commandline << " " << array[i];
    }

    Message( prefix, "\n>>>>>>>>>>>>>>>>>> TRACE <<<<<<<<<<<<<<<<\n",0,0);

    //out << commandline.str() << endl;

    system( commandline.str().c_str() );

}

/* ================================================================== */
string StringFromDouble(double val , UINT32 precision, UINT32 width)
{
    ostringstream ostr;
    ostr.setf(ios::fixed,ios::floatfield);
    ostr << setw(width) << setprecision(precision) << val;
    return ostr.str();
}

UINT32 Uint32FromString(const string& s)
{
    return strtoul( s.c_str(),0,0);
}


INT32 Int32FromString(const string& s)
{
    return strtol( s.c_str(),0,0);
}

/* ================================================================== */
/*! 
  convert an INT32 into a dec string. Padding can be specified as well   
 */

string StringDecFromInt32( INT32 l, UINT32 digits, CHAR padding)
{
  
    const BOOL negative = (l < 0);
    
    if( l < 0 ) l = -l;  // FIXME: avoid bug for negative num that cannot be negated

    CHAR buffer[32];

    UINT32 i = 31;
    buffer[i] = '\0';
    
    do
    {
        i--;
        buffer[i] = '0' + l % 10;
        l /= 10;
    }while(l);

    if( negative )
    {
        i--;
        buffer[i] = '-';
    }


    // FIXME bug minus signed should follow after padding
    
    digits = sizeof(buffer) - digits - 1;
    while( i > digits )
    {
        i--;
        buffer[i] = padding;
    }
    
    return string(&buffer[i]);
}

/* ================================================================== */
VOID SignalHandler(int arg)
{
    string prefix("X: ");

    Message( prefix, ">>>>>>>>>>>>>>>>>> SIGNAL CAUGHT <<<<<<<<<<<<<<<<\n",1,1);    
    Message(prefix,"Signal caught\n",1,1);
}

/* ================================================================== */

VOID InstallSignalHandlers(char *name)
{
    image_name = name;
    
    signal( 8, SignalHandler);
    signal(10, SignalHandler);
    signal(11, SignalHandler);    
}

/*  ================================================================== */

string Quote(const string& is)
{
    string os("\"");
    
    const UINT32 length = is.length();
    for (UINT32 i=0; i < length; ++i )
    {
        if( is[i] == '\"' ) os.append(1,'\\');

        os.append(1,is[i]);
    }

    os.append(1,'\"');

    
    
    return os;
}

/*  ================================================================== */

string Unquote(const string& is)
{
    string os;

    const UINT32 length = is.length();

    for (UINT32 i=0; i < length; ++i )
    {
        if( is[i] == '\"' )
        {
            ASSERTX( i==0 || i== length-1);
            continue;
        }
        
        if( is[i] == '\\' )
        {
            ++i;
            ASSERTX( i < length);
        }
        
        os.append(1,is[i]);
    }

    return os;
}


/*  ================================================================== */

INT32 Tokenize(const string& line, string *array, UINT32 n)
{

    const UINT32 length = line.length();

    UINT32 pos_start = 0;
    UINT32 count = 0;
    
    for( ; count < n; count++ )
    {
        BOOL quoted_string = 0;

        // skip leading white space
        while( pos_start < length )
        {
            const CHAR c = line[pos_start];
                        
            if( isspace(c) )
            {
                pos_start++;
            }
            else
            {
                if ( c == '\"' )
                {
                    quoted_string = 1;
                }
                
                break;
            }
        }
            
        if( pos_start >= length )
            break;

        //find toked end

        UINT32 pos_end = pos_start;

        if ( quoted_string )
            ++pos_end;
        
        while( pos_end < length )
        {
            const CHAR c = line[pos_end];
            
            if( !quoted_string && isspace(c) )
            {
                break;
            }
            else if( quoted_string && c == '\"' )
            {
                pos_end++;
                quoted_string = 0;
                break;
            }
            else if( c == '\\')
            {
                pos_end+=2;
                if ( pos_end > length ) return -1;
            }
            else
            {
                pos_end++;
            }
        }

        if ( quoted_string ) return -1;
        
        array[count] = line.substr(pos_start,pos_end -pos_start);
        
        pos_start = pos_end;
    }
    
    return count;
}

/* ================================================================== */

#ifdef STAND_ALONE

void cool()
{
    DumpTrace();
}

int main(int argc, char *argv[])
{
    InstallSignalHandlers(argv[0]);
    
    cool();
    
    return 0;
}

#endif
